#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Any, Callable, Dict, List, Optional, Type, Union

import torch
from torch import nn

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import TEXT
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.reduction_modules import SequenceReducer
from ludwig.schema.encoders.sequence_encoders import SequenceEncoderConfig
from ludwig.schema.encoders.text_encoders import (
    ALBERTConfig,
    AutoTransformerConfig,
    BERTConfig,
    CamemBERTConfig,
    CTRLConfig,
    DistilBERTConfig,
    ELECTRAConfig,
    FlauBERTConfig,
    GPT2Config,
    GPTConfig,
    LongformerConfig,
    MT5Config,
    RoBERTaConfig,
    T5Config,
    TransformerXLConfig,
    XLMConfig,
    XLMRoBERTaConfig,
    XLNetConfig,
)
from ludwig.utils.hf_utils import load_pretrained_hf_model
from ludwig.utils.torch_utils import FreezeModule

logger = logging.getLogger(__name__)


def _cls_pooled_error_message(encoder: str):
    # TODO(Arnav): Remove this once we have reduce_output options set for
    # each encoder type in the schema
    raise ValueError(f"reduce_output cannot be cls_pooled for {encoder}")


class HFTextEncoder(Encoder):
    DEFAULT_MODEL_NAME: str

    def _init_config(self, transformer, schema_keys: List[str], encoder_config: SequenceEncoderConfig):
        """Creates a config object for the encoder using the transformer model and the passed-in encoder config.

        The transformer's config is only known after it is instantiated, so we must update the
        encoder config with the values from the transformer config.

        Args:
            transformer: The transformer model.
            schema_keys: The keys in the encoder config schema. We only want to update the encoder config
                with the values from the transformer config that are in the schema.
            encoder_config: The existing encoder config containing defaults and user-specified values.
                If the values in this config differ from the transformer's config, the transformer's config
                values will override this config's values.
        Returns:
            A new encoder config object with the updated values from the transformer config.
        """
        transformer_config = transformer.config.to_dict()
        final_hf_config_params = {k: v for k, v in transformer_config.items() if k in schema_keys}
        encoder_config_dict = encoder_config.to_dict()
        encoder_config_dict.update(final_hf_config_params)
        return self.get_schema_cls().from_dict(encoder_config_dict)

    def _init_transformer_from_scratch(
        self, hf_model_cls: Type, hf_config_cls: Type, hf_config_params: Dict[str, Any], vocab_size: int
    ):
        """Initializes the transformer model from scratch. This is in contrast to loading a pre-trained model.

        Args:
            hf_model_cls: The HuggingFace model class.
            hf_config_cls: The HuggingFace config class.
            hf_config_params: The HuggingFace config parameters exposed through the Ludwig schema.
            vocab_size: The vocab size of the dataset. Because we are training from scratch, we can resize the
                token embeddings table freely.
        Returns:
            The transformer model.
        """
        config = hf_config_cls(**hf_config_params)
        transformer = hf_model_cls(config)
        self._maybe_resize_token_embeddings(transformer, vocab_size)
        return transformer

    def _maybe_resize_token_embeddings(self, transformer, vocab_size: int):
        """Resizes the token embeddings if the vocab size is different from the transformer's vocab size.

        This should only happen if we are instantiating a model from scratch (i.e. not loading from a pretrained model
        or checkpoint). Pretrained models update the vocab size stored in the config. This means if we are loading a
        pretrained model from a checkpoint, the config vocab size should match the model's vocab size.

        It is important that pretrained models update the vocab size stored in the config because sometimes the
        pretrained models will have an embeddings table that is a different size than the vocab size. Examples:

        CamemBERT:  https://github.com/huggingface/tokenizers/issues/900#issue-1122256698
        T5:         https://github.com/huggingface/transformers/issues/4875#issue-635471552

        Args:
            transformer: The transformer model.
            vocab_size: The vocab size of the dataset.
        """
        if vocab_size != transformer.config.vocab_size:
            transformer.resize_token_embeddings(vocab_size)

    def get_embedding_layer(self) -> nn.Module:
        return next(self.transformer.module.children())


@DeveloperAPI
@register_encoder("albert", TEXT)
class ALBERTEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "albert-base-v2"

    def __init__(
        self,
        max_sequence_length,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        trainable: bool = False,
        reduce_output: str = "cls_pooled",
        vocab_size: int = 30000,
        embedding_size: int = 128,
        hidden_size: int = 4096,
        num_hidden_layers: int = 12,
        num_hidden_groups: int = 1,
        num_attention_heads: int = 64,
        intermediate_size: int = 16384,
        inner_group_num: int = 1,
        hidden_act: str = "gelu_new",
        hidden_dropout_prob: float = 0,
        attention_probs_dropout_prob: float = 0,
        max_position_embeddings: int = 512,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        classifier_dropout_prob: float = 0.1,
        position_embedding_type: str = "absolute",
        pad_token_id: int = 0,
        bos_token_id: int = 2,
        eos_token_id: int = 3,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import AlbertConfig, AlbertModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            embedding_size=embedding_size,
            hidden_size=hidden_size,
            num_hidden_layers=num_hidden_layers,
            num_hidden_groups=num_hidden_groups,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            inner_group_num=inner_group_num,
            hidden_act=hidden_act,
            hidden_dropout_prob=hidden_dropout_prob,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            max_position_embeddings=max_position_embeddings,
            type_vocab_size=type_vocab_size,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            classifier_dropout_prob=classifier_dropout_prob,
            position_embedding_type=position_embedding_type,
            pad_token_id=pad_token_id,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(AlbertModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(AlbertModel, AlbertConfig, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if not self.reduce_output == "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            hidden = transformer_outputs[1]
        else:
            hidden = transformer_outputs[0][:, 1:-1, :]
            hidden = self.reduce_sequence(hidden, self.reduce_output)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return ALBERTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by BERT tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("mt5", TEXT)
class MT5Encoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "google/mt5-base"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        trainable: bool = False,
        reduce_output: str = "sum",
        vocab_size: int = 250112,
        d_model: int = 512,
        d_kv: int = 64,
        d_ff: int = 1024,
        num_layers: int = 8,
        num_decoder_layers: int = None,
        num_heads: int = 6,
        relative_attention_num_buckets: int = 32,
        dropout_rate: float = 0.1,
        layer_norm_epsilon: float = 1e-06,
        initializer_factor: float = 1.0,
        feed_forward_proj: str = "gated-gelu",
        is_encoder_decoder: bool = True,
        use_cache: bool = True,
        tokenizer_class: str = "T5Tokenizer",
        tie_word_embeddings: bool = False,
        pad_token_id: int = 0,
        eos_token_id: int = 1,
        decoder_start_token_id: int = 0,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import MT5Config, MT5EncoderModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            d_model=d_model,
            d_kv=d_kv,
            d_ff=d_ff,
            num_layers=num_layers,
            num_decoder_layers=num_decoder_layers,
            num_heads=num_heads,
            relative_attention_num_buckets=relative_attention_num_buckets,
            dropout_rate=dropout_rate,
            layer_norm_epsilon=layer_norm_epsilon,
            initializer_factor=initializer_factor,
            feed_forward_proj=feed_forward_proj,
            is_encoder_decoder=is_encoder_decoder,
            use_cache=use_cache,
            tokenizer_class=tokenizer_class,
            tie_word_embeddings=tie_word_embeddings,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            decoder_start_token_id=decoder_start_token_id,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(MT5EncoderModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(MT5EncoderModel, MT5Config, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
        )
        hidden = transformer_outputs[0][:, 1:-1, :]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return MT5Config

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by MT5 tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("xlmroberta", TEXT)
class XLMRoBERTaEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "xlm-roberta-base"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "cls_pooled",
        trainable: bool = False,
        vocab_size: int = None,
        pad_token_id: int = 1,
        bos_token_id: int = 0,
        eos_token_id: int = 2,
        max_position_embeddings: int = 514,
        type_vocab_size: int = 1,
        add_pooling_layer: bool = True,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import XLMRobertaConfig, XLMRobertaModel

        hf_config_params = dict(
            pad_token_id=pad_token_id,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            max_position_embeddings=max_position_embeddings,
            type_vocab_size=type_vocab_size,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(XLMRobertaModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(
                XLMRobertaModel, XLMRobertaConfig, hf_config_params, vocab_size
            )

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if not self.reduce_output == "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            hidden = transformer_outputs[1]
        else:
            hidden = transformer_outputs[0][:, 1:-1, :]
            hidden = self.reduce_sequence(hidden, self.reduce_output)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return XLMRoBERTaConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by XLMRoberta tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("bert", TEXT)
class BERTEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "bert-base-uncased"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        trainable: bool = False,
        reduce_output: str = "cls_pooled",
        vocab_size: int = 30522,
        hidden_size: int = 768,
        num_hidden_layers: int = 12,
        num_attention_heads: int = 12,
        intermediate_size: int = 3072,
        hidden_act: Union[str, Callable] = "gelu",
        hidden_dropout_prob: float = 0.1,
        attention_probs_dropout_prob: float = 0.1,
        max_position_embeddings: int = 512,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        pad_token_id: int = 0,
        gradient_checkpointing: bool = False,
        position_embedding_type: str = "absolute",
        classifier_dropout: float = None,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import BertConfig, BertModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            hidden_size=hidden_size,
            num_hidden_layers=num_hidden_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            hidden_act=hidden_act,
            hidden_dropout_prob=hidden_dropout_prob,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            max_position_embeddings=max_position_embeddings,
            type_vocab_size=type_vocab_size,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            pad_token_id=pad_token_id,
            gradient_checkpointing=gradient_checkpointing,
            position_embedding_type=position_embedding_type,
            classifier_dropout=classifier_dropout,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(BertModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(BertModel, BertConfig, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if not self.reduce_output == "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)

        self.transformer = FreezeModule(transformer, frozen=not trainable)

        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            hidden = transformer_outputs[1]
        else:
            hidden = transformer_outputs[0][:, 1:-1, :]
            hidden = self.reduce_sequence(hidden, self.reduce_output)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return BERTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    # TODO(shreya): Confirm that this is it
    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by BERT tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("xlm", TEXT)
class XLMEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "xlm-mlm-en-2048"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        trainable: bool = False,
        reduce_output: str = "sum",
        vocab_size: int = 30145,
        emb_dim: int = 2048,
        n_layers: int = 12,
        n_heads: int = 16,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        gelu_activation: bool = True,
        sinusoidal_embeddings: bool = False,
        causal: bool = False,
        asm: bool = False,
        n_langs: int = 1,
        use_lang_emb: bool = True,
        max_position_embeddings: int = 512,
        embed_init_std: float = 2048**-0.5,
        layer_norm_eps: float = 1e-12,
        init_std: float = 0.02,
        bos_index: int = 0,
        eos_index: int = 1,
        pad_index: int = 2,
        unk_index: int = 3,
        mask_index: int = 5,
        is_encoder: bool = True,
        start_n_top: int = 5,
        end_n_top: int = 5,
        mask_token_id: int = 0,
        lang_id: int = 0,
        pad_token_id: int = 2,
        bos_token_id: int = 0,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import XLMConfig, XLMModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            emb_dim=emb_dim,
            n_layers=n_layers,
            n_heads=n_heads,
            dropout=dropout,
            attention_dropout=attention_dropout,
            gelu_activation=gelu_activation,
            sinusoidal_embeddings=sinusoidal_embeddings,
            causal=causal,
            asm=asm,
            n_langs=n_langs,
            use_lang_emb=use_lang_emb,
            max_position_embeddings=max_position_embeddings,
            embed_init_std=embed_init_std,
            layer_norm_eps=layer_norm_eps,
            init_std=init_std,
            bos_index=bos_index,
            eos_index=eos_index,
            pad_index=pad_index,
            unk_index=unk_index,
            mask_index=mask_index,
            is_encoder=is_encoder,
            start_n_top=start_n_top,
            end_n_top=end_n_top,
            mask_token_id=mask_token_id,
            lang_id=lang_id,
            pad_token_id=pad_token_id,
            bos_token_id=bos_token_id,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(XLMModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(XLMModel, XLMConfig, hf_config_params, vocab_size)

        self.config = self._init_config(transformer, hf_config_params, encoder_config)

        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return XLMConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    # TODO(shreya): Confirm that this is it
    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by BERT tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("gpt", TEXT)
class GPTEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "openai-gpt"

    def __init__(
        self,
        max_sequence_length: int,
        reduce_output: str = "sum",
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        trainable: bool = False,
        vocab_size: int = 30522,
        n_positions: int = 40478,
        n_ctx: int = 512,
        n_embd: int = 768,
        n_layer: int = 12,
        n_head: int = 12,
        afn: str = "gelu",
        resid_pdrop: float = 0.1,
        embd_pdrop: float = 0.1,
        attn_pdrop: float = 0.1,
        layer_norm_epsilon: float = 1e-5,
        initializer_range: float = 0.02,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import OpenAIGPTConfig, OpenAIGPTModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            n_positions=n_positions,
            n_ctx=n_ctx,
            n_embd=n_embd,
            n_layer=n_layer,
            n_head=n_head,
            afn=afn,
            resid_pdrop=resid_pdrop,
            embd_pdrop=embd_pdrop,
            attn_pdrop=attn_pdrop,
            layer_norm_epsilon=layer_norm_epsilon,
            initializer_range=initializer_range,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(OpenAIGPTModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(
                OpenAIGPTModel, OpenAIGPTConfig, hf_config_params, vocab_size
            )

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return GPTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return torch.Size([self.max_sequence_length, self.transformer.module.config.hidden_size])
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("gpt2", TEXT)
class GPT2Encoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "gpt2"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 50257,
        n_positions: int = 1024,
        n_ctx: int = 1024,
        n_embd: int = 768,
        n_layer: int = 12,
        n_head: int = 12,
        n_inner: Optional[int] = None,
        activation_function: str = "gelu",
        resid_pdrop: float = 0.1,
        embd_pdrop: float = 0.1,
        attn_pdrop: float = 0.1,
        layer_norm_epsilon: float = 1e-5,
        initializer_range: float = 0.02,
        scale_attn_weights: bool = True,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import GPT2Config, GPT2Model

        hf_config_params = dict(
            vocab_size=vocab_size,
            n_positions=n_positions,
            n_ctx=n_ctx,
            n_embd=n_embd,
            n_layer=n_layer,
            n_head=n_head,
            n_inner=n_inner,
            activation_function=activation_function,
            resid_pdrop=resid_pdrop,
            embd_pdrop=embd_pdrop,
            attn_pdrop=attn_pdrop,
            layer_norm_epsilon=layer_norm_epsilon,
            initializer_range=initializer_range,
            scale_attn_weights=scale_attn_weights,
        )

        if use_pretrained:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(GPT2Model, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(GPT2Model, GPT2Config, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return GPT2Config

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return torch.Size([self.max_sequence_length, self.transformer.module.config.hidden_size])
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("roberta", TEXT)
class RoBERTaEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "roberta-base"

    def __init__(
        self,
        max_sequence_length,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "cls_pooled",
        trainable: bool = False,
        vocab_size: int = None,
        pad_token_id: int = 1,
        bos_token_id: int = 0,
        eos_token_id: int = 2,
        max_position_embeddings: int = 514,
        type_vocab_size: int = 1,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import RobertaConfig, RobertaModel

        hf_config_params = dict(
            pad_token_id=pad_token_id,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            max_position_embeddings=max_position_embeddings,
            type_vocab_size=type_vocab_size,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(RobertaModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(RobertaModel, RobertaConfig, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length
        self.reduce_output = reduce_output
        if not self.reduce_output == "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            hidden = transformer_outputs[1]
        else:
            hidden = transformer_outputs[0][:, 1:-1, :]  # bos + [sent] + sep
            hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return RoBERTaConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return torch.Size([self.max_sequence_length - 2, self.transformer.module.config.hidden_size])
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("transformer_xl", TEXT)
class TransformerXLEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "transfo-xl-wt103"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 267735,
        cutoffs: List[int] = [20000, 40000, 200000],
        d_model: int = 1024,
        d_embed: int = 1024,
        n_head: int = 16,
        d_head: int = 64,
        d_inner: int = 4096,
        div_val: int = 4,
        pre_lnorm: bool = False,
        n_layer: int = 18,
        mem_len: int = 1600,
        clamp_len: int = 1000,
        same_length: bool = True,
        proj_share_all_but_first: bool = True,
        attn_type: int = 0,
        sample_softmax: int = -1,
        adaptive: bool = True,
        dropout: float = 0.1,
        dropatt: float = 0.0,
        untie_r: bool = True,
        init: str = "normal",
        init_range: float = 0.01,
        proj_init_std: float = 0.01,
        init_std: float = 0.02,
        layer_norm_epsilon: float = 1e-5,
        eos_token_id: int = 0,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import TransfoXLConfig, TransfoXLModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            cutoffs=cutoffs,
            d_model=d_model,
            d_embed=d_embed,
            n_head=n_head,
            d_head=d_head,
            d_inner=d_inner,
            div_val=div_val,
            pre_lnorm=pre_lnorm,
            n_layer=n_layer,
            mem_len=mem_len,
            clamp_len=clamp_len,
            same_length=same_length,
            proj_share_all_but_first=proj_share_all_but_first,
            attn_type=attn_type,
            sample_softmax=sample_softmax,
            adaptive=adaptive,
            dropout=dropout,
            dropatt=dropatt,
            untie_r=untie_r,
            init=init,
            init_range=init_range,
            proj_init_std=proj_init_std,
            init_std=init_std,
            layer_norm_epsilon=layer_norm_epsilon,
            eos_token_id=eos_token_id,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(TransfoXLModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            config = TransfoXLConfig(**hf_config_params)
            transformer = TransfoXLModel(config)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        transformer_outputs = self.transformer.module(inputs)
        hidden = transformer_outputs[0]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return TransformerXLConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return torch.Size([self.max_sequence_length, self.transformer.module.config.d_model])
        else:
            return torch.Size([self.transformer.module.config.d_model])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("xlnet", TEXT)
class XLNetEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "xlnet-base-cased"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 32000,
        d_model: int = 1024,
        n_layer: int = 24,
        n_head: int = 16,
        d_inner: int = 4096,
        ff_activation: str = "gelu",
        untie_r: bool = True,
        attn_type: str = "bi",
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        dropout: float = 0.1,
        mem_len: Optional[int] = 512,
        reuse_len: Optional[int] = None,
        use_mems_eval: bool = True,
        use_mems_train: bool = False,
        bi_data: bool = False,
        clamp_len: int = -1,
        same_length: bool = False,
        summary_type: str = "last",
        summary_use_proj: bool = True,
        summary_activation: str = "tanh",
        summary_last_dropout: float = 0.1,
        start_n_top: int = 5,
        end_n_top: int = 5,
        pad_token_id: int = 5,
        bos_token_id: int = 1,
        eos_token_id: int = 2,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import XLNetConfig, XLNetModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            d_model=d_model,
            n_layer=n_layer,
            n_head=n_head,
            d_inner=d_inner,
            ff_activation=ff_activation,
            untie_r=untie_r,
            attn_type=attn_type,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            mem_len=mem_len,
            reuse_len=reuse_len,
            use_mems_eval=use_mems_eval,
            use_mems_train=use_mems_train,
            bi_data=bi_data,
            clamp_len=clamp_len,
            same_length=same_length,
            summary_type=summary_type,
            summary_use_proj=summary_use_proj,
            summary_activation=summary_activation,
            summary_last_dropout=summary_last_dropout,
            start_n_top=start_n_top,
            end_n_top=end_n_top,
            pad_token_id=pad_token_id,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(XLNetModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(XLNetModel, XLNetConfig, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.max_sequence_length = max_sequence_length
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)

    def forward(self, inputs: torch.Tensor, mask: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return XLNetConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return torch.Size([self.max_sequence_length, self.transformer.module.config.d_model])
        else:
            return torch.Size([self.transformer.module.config.d_model])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("distilbert", TEXT)
class DistilBERTEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "distilbert-base-uncased"

    def __init__(
        self,
        max_sequence_length: int,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        use_pretrained: bool = True,
        vocab_size: int = 30522,
        max_position_embeddings: int = 512,
        sinusoidal_pos_embds: bool = False,
        n_layers: int = 6,
        n_heads: int = 12,
        dim: int = 768,
        hidden_dim: int = 3072,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        activation: Union[str, Callable] = "gelu",
        initializer_range: float = 0.02,
        qa_dropout: float = 0.1,
        seq_classif_dropout: float = 0.2,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import DistilBertConfig, DistilBertModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            max_position_embeddings=max_position_embeddings,
            sinusoidal_pos_embds=sinusoidal_pos_embds,
            n_layers=n_layers,
            n_heads=n_heads,
            dim=dim,
            hidden_dim=hidden_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            activation=activation,
            initializer_range=initializer_range,
            qa_dropout=qa_dropout,
            seq_classif_dropout=seq_classif_dropout,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(DistilBertModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(
                DistilBertModel, DistilBertConfig, hf_config_params, vocab_size
            )

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.max_sequence_length = max_sequence_length
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.last_inputs = None
        self.last_hidden = None

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)

        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
        )
        hidden = transformer_outputs[0][:, 1:-1, :]
        self.last_inputs = inputs
        self.last_hidden = hidden
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return DistilBERTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by BERT tokenizer.
            return torch.Size([self.max_sequence_length - 2, self.transformer.module.config.dim])
        return torch.Size([self.transformer.module.config.dim])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("ctrl", TEXT)
class CTRLEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "ctrl"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 246534,
        n_positions: int = 256,
        n_ctx: int = 256,
        n_embd: int = 1280,
        dff: int = 8192,
        n_layer: int = 48,
        n_head: int = 16,
        resid_pdrop: float = 0.1,
        embd_pdrop: float = 0.1,
        attn_pdrop: float = 0.1,
        layer_norm_epsilon: float = 1e-6,
        initializer_range: float = 0.02,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import CTRLConfig, CTRLModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            n_positions=n_positions,
            n_ctx=n_ctx,
            n_embd=n_embd,
            dff=dff,
            n_layer=n_layer,
            n_head=n_head,
            resid_pdrop=resid_pdrop,
            embd_pdrop=embd_pdrop,
            attn_pdrop=attn_pdrop,
            layer_norm_epsilon=layer_norm_epsilon,
            initializer_range=initializer_range,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(CTRLModel, pretrained_model_name_or_path, **pretrained_kwargs)
            self.vocab_size = transformer.config.vocab_size
        else:
            transformer = self._init_transformer_from_scratch(CTRLModel, CTRLConfig, hf_config_params, vocab_size)
            self.vocab_size = vocab_size

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.max_sequence_length = max_sequence_length
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return CTRLConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return torch.Size([self.max_sequence_length, self.transformer.module.config.n_embd])
        return torch.Size([self.transformer.module.config.n_embd])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("camembert", TEXT)
class CamemBERTEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "camembert-base"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "cls-pooled",
        trainable: bool = False,
        vocab_size: int = 30522,
        hidden_size: int = 768,
        num_hidden_layers: int = 12,
        num_attention_heads: int = 12,
        intermediate_size: int = 3072,
        hidden_act: Union[str, Callable] = "gelu",
        hidden_dropout_prob: float = 0.1,
        attention_probs_dropout_prob: float = 0.1,
        max_position_embeddings: int = 512,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        pad_token_id: int = 0,
        gradient_checkpointing: bool = False,
        position_embedding_type: str = "absolute",
        classifier_dropout: float = None,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import CamembertConfig, CamembertModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            hidden_size=hidden_size,
            num_hidden_layers=num_hidden_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            hidden_act=hidden_act,
            hidden_dropout_prob=hidden_dropout_prob,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            max_position_embeddings=max_position_embeddings,
            type_vocab_size=type_vocab_size,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            pad_token_id=pad_token_id,
            gradient_checkpointing=gradient_checkpointing,
            position_embedding_type=position_embedding_type,
            classifier_dropout=classifier_dropout,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(CamembertModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(
                CamembertModel, CamembertConfig, hf_config_params, vocab_size
            )

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.reduce_output = reduce_output
        if not self.reduce_output == "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            hidden = transformer_outputs[1]
        else:
            hidden = transformer_outputs[0][:, 1:-1, :]
            hidden = self.reduce_sequence(hidden, self.reduce_output)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return CamemBERTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by BERT tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("t5", TEXT)
class T5Encoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "t5-small"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 32128,
        d_model: int = 512,
        d_kv: int = 64,
        d_ff: int = 2048,
        num_layers: int = 6,
        num_decoder_layers: Optional[int] = None,
        num_heads: int = 8,
        relative_attention_num_buckets: int = 32,
        dropout_rate: float = 0.1,
        layer_norm_eps: float = 1e-6,
        initializer_factor: float = 1,
        feed_forward_proj: str = "relu",
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import T5Config, T5Model

        hf_config_params = dict(
            vocab_size=vocab_size,
            d_model=d_model,
            d_kv=d_kv,
            d_ff=d_ff,
            num_layers=num_layers,
            num_decoder_layers=num_decoder_layers,
            num_heads=num_heads,
            relative_attention_num_buckets=relative_attention_num_buckets,
            dropout_rate=dropout_rate,
            layer_norm_eps=layer_norm_eps,
            initializer_factor=initializer_factor,
            feed_forward_proj=feed_forward_proj,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(T5Model, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(T5Model, T5Config, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.max_sequence_length = max_sequence_length
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            inputs,
            decoder_input_ids=inputs,
            attention_mask=mask,
        )
        hidden = transformer_outputs[0][:, 0:-1, :]  # [eos token]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return T5Config

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 1 to remove EOS token added by T5 tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 1,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.d_model])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("flaubert", TEXT)
class FlauBERTEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "flaubert/flaubert_small_cased"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 30145,
        pre_norm: bool = False,
        layerdrop: float = 0.0,
        emb_dim: int = 2048,
        n_layers: int = 12,
        n_heads: int = 16,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        gelu_activation: bool = True,
        sinusoidal_embeddings: bool = False,
        causal: bool = False,
        asm: bool = False,
        n_langs: int = 1,
        use_lang_emb: bool = True,
        max_position_embeddings: int = 512,
        embed_init_std: float = 2048**-0.5,
        init_std: int = 0.02,
        layer_norm_eps: float = 1e-12,
        bos_index: int = 0,
        eos_index: int = 1,
        pad_index: int = 2,
        unk_index: int = 3,
        mask_index: int = 5,
        is_encoder: bool = True,
        mask_token_id: int = 0,
        lang_id: int = 1,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import FlaubertConfig, FlaubertModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            pre_norm=pre_norm,
            layerdrop=layerdrop,
            emb_dim=emb_dim,
            n_layers=n_layers,
            n_heads=n_heads,
            dropout=dropout,
            attention_dropout=dropout,
            gelu_activation=gelu_activation,
            sinusoidal_embeddings=sinusoidal_embeddings,
            causal=causal,
            asm=asm,
            n_langs=n_langs,
            use_lang_emb=use_lang_emb,
            max_position_embeddings=max_position_embeddings,
            embed_init_std=embed_init_std,
            init_std=init_std,
            layer_norm_eps=layer_norm_eps,
            bos_index=bos_index,
            eos_index=eos_index,
            pad_index=pad_index,
            unk_index=unk_index,
            mask_index=mask_index,
            is_encoder=is_encoder,
            mask_token_id=mask_token_id,
            lang_id=lang_id,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(FlaubertModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(
                FlaubertModel, FlaubertConfig, hf_config_params, vocab_size
            )

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.max_sequence_length = max_sequence_length
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0][:, 1:-1, :]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return FlauBERTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.emb_dim])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("electra", TEXT)
class ELECTRAEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "google/electra-small-discriminator"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: int = 30522,
        embedding_size: int = 128,
        hidden_size: int = 256,
        num_hidden_layers: int = 12,
        num_attention_heads: int = 4,
        intermediate_size: int = 1024,
        hidden_act: Union[str, Callable] = "gelu",
        hidden_dropout_prob: float = 0.1,
        attention_probs_dropout_prob: float = 0.1,
        max_position_embeddings: int = 512,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        position_embedding_type: str = "absolute",
        classifier_dropout: Optional[float] = None,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import ElectraConfig, ElectraModel

        hf_config_params = dict(
            vocab_size=vocab_size,
            embedding_size=embedding_size,
            hidden_size=hidden_size,
            num_hidden_layers=num_hidden_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            hidden_act=hidden_act,
            hidden_dropout_prob=hidden_dropout_prob,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            max_position_embeddings=max_position_embeddings,
            type_vocab_size=type_vocab_size,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            position_embedding_type=position_embedding_type,
            classifier_dropout=classifier_dropout,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(ElectraModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(ElectraModel, ElectraConfig, hf_config_params, vocab_size)

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.max_sequence_length = max_sequence_length
        self.reduce_output = reduce_output
        if self.reduce_output == "cls_pooled":
            _cls_pooled_error_message(self.__class__.__name__)
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> Dict[str, torch.Tensor]:
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        hidden = transformer_outputs[0][:, 1:-1, :]
        hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return ELECTRAConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("longformer", TEXT)
class LongformerEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = "allenai/longformer-base-4096"

    def __init__(
        self,
        max_sequence_length: int,
        use_pretrained: bool = True,
        attention_window: Union[List[int], int] = 512,
        sep_token_id: int = 2,
        pretrained_model_name_or_path: str = DEFAULT_MODEL_NAME,
        saved_weights_in_checkpoint: bool = False,
        reduce_output: Optional[str] = "cls_pooled",
        trainable: bool = False,
        vocab_size: int = 50265,
        num_tokens: Optional[int] = None,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import LongformerConfig, LongformerModel

        hf_config_params = dict(
            attention_window=attention_window,
            sep_token_id=sep_token_id,
            vocab_size=vocab_size,
            **kwargs,
        )

        if use_pretrained and not saved_weights_in_checkpoint:
            pretrained_kwargs = pretrained_kwargs or {}
            transformer = load_pretrained_hf_model(LongformerModel, pretrained_model_name_or_path, **pretrained_kwargs)
        else:
            transformer = self._init_transformer_from_scratch(
                LongformerModel, LongformerConfig, hf_config_params, vocab_size
            )

        if encoder_config is not None:
            self.config = self._init_config(transformer, hf_config_params.keys(), encoder_config)
        else:
            self.config = None

        self.reduce_output = reduce_output
        if not self.reduce_output == "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.max_sequence_length = max_sequence_length

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            hidden = transformer_outputs[1]
        else:
            hidden = transformer_outputs[0][:, 1:-1, :]  # bos + [sent] + sep
            hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return LongformerConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # Subtract 2 to remove CLS and PAD tokens added by Longformer (== Roberta) tokenizer.
            return torch.Size(
                [
                    self.max_sequence_length - 2,
                    self.transformer.module.config.hidden_size,
                ]
            )
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32


@DeveloperAPI
@register_encoder("auto_transformer", TEXT)
class AutoTransformerEncoder(HFTextEncoder):
    DEFAULT_MODEL_NAME = None

    def __init__(
        self,
        pretrained_model_name_or_path: str,
        max_sequence_length: int,
        reduce_output: str = "sum",
        trainable: bool = False,
        vocab_size: Optional[int] = None,
        pretrained_kwargs: Dict = None,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()

        from transformers import AutoModel

        pretrained_kwargs = pretrained_kwargs or {}
        transformer = load_pretrained_hf_model(AutoModel, pretrained_model_name_or_path, **pretrained_kwargs)
        self._maybe_resize_token_embeddings(transformer, vocab_size)

        self.config = self._init_config(transformer, [], encoder_config)

        self.transformer = FreezeModule(transformer, frozen=not trainable)
        self.reduce_output = reduce_output
        if self.reduce_output != "cls_pooled":
            self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)
        self.max_sequence_length = max_sequence_length

    def _maybe_resize_token_embeddings(self, transformer, vocab_size: Optional[int] = None):
        """Overridden because AutoModel should use its own vocab size unless vocab size is explicitly specified."""
        if vocab_size is not None:
            transformer.resize_token_embeddings(vocab_size)
            self.vocab_size = vocab_size
        else:
            self.vocab_size = transformer.config.vocab_size

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        if mask is not None:
            mask = mask.to(torch.int32)
        transformer_outputs = self.transformer.module(
            input_ids=inputs,
            attention_mask=mask,
            token_type_ids=torch.zeros_like(inputs),
        )
        if self.reduce_output == "cls_pooled":
            # this works only if the user know that the specific model
            # they want to use has the same outputs of
            # the BERT base class call() function
            hidden = transformer_outputs["pooler_output"]
        else:
            hidden = transformer_outputs["last_hidden_state"]
            hidden = self.reduce_sequence(hidden, self.reduce_output)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return AutoTransformerConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            # TODO(justin): This may need to be conditioned on which AutoModel gets chosen.
            return torch.Size([self.max_sequence_length, self.transformer.module.config.hidden_size])
        return torch.Size([self.transformer.module.config.hidden_size])

    @property
    def input_dtype(self):
        return torch.int32



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import contextlib
import copy
import logging
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TYPE_CHECKING, Union

import dask
import numpy as np
import pandas as pd
import ray
import torch
import tqdm
from packaging import version
from pyarrow.fs import FSSpecHandler, PyFileSystem
from ray import ObjectRef
from ray.air import session
from ray.air.checkpoint import Checkpoint
from ray.air.config import DatasetConfig, RunConfig, ScalingConfig
from ray.train.torch import TorchCheckpoint
from ray.util.dask import ray_dask_get
from ray.util.placement_group import placement_group, remove_placement_group

from ludwig.distributed import get_current_dist_strategy, get_default_strategy_name, get_dist_strategy
from ludwig.utils.batch_size_tuner import BatchSizeEvaluator

if TYPE_CHECKING:
    from ludwig.api import LudwigModel

from ludwig.api_annotations import DeveloperAPI
from ludwig.backend.base import Backend, RemoteTrainingMixin
from ludwig.backend.datasource import BinaryIgnoreNoneTypeDatasource
from ludwig.constants import CPU_RESOURCES_PER_TRIAL, EXECUTOR, MODEL_ECD, NAME, PROC_COLUMN, TYPE
from ludwig.data.dataframe.base import DataFrameEngine
from ludwig.data.dataframe.dask import tensor_extension_casting
from ludwig.data.dataset.ray import _SCALAR_TYPES, RayDataset, RayDatasetManager, RayDatasetShard
from ludwig.models.base import BaseModel
from ludwig.models.ecd import ECD
from ludwig.models.predictor import BasePredictor, get_output_columns, Predictor, RemotePredictor
from ludwig.schema.trainer import ECDTrainerConfig
from ludwig.trainers.registry import ray_trainers_registry, register_ray_trainer
from ludwig.trainers.trainer import BaseTrainer, RemoteTrainer
from ludwig.types import HyperoptConfigDict, ModelConfigDict, TrainerConfigDict, TrainingSetMetadataDict
from ludwig.utils.dataframe_utils import set_index_name
from ludwig.utils.fs_utils import get_fs_and_path
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.system_utils import Resources
from ludwig.utils.torch_utils import get_torch_device, initialize_pytorch
from ludwig.utils.types import DataFrame, Series

_ray230 = version.parse(ray.__version__) >= version.parse("2.3.0")


logger = logging.getLogger(__name__)


RAY_DEFAULT_PARALLELISM = 200
FIFTEEN_MINS_IN_S = 15 * 60


def _num_nodes() -> int:
    node_resources = [node["Resources"] for node in ray.nodes()]
    return len(node_resources)


def get_trainer_kwargs(**kwargs) -> TrainerConfigDict:
    kwargs = copy.deepcopy(kwargs)

    # Our goal is to have a worker per resource used for training.
    # The priority is GPUs, but can fall back to CPUs if there are no
    # GPUs available.
    use_gpu = kwargs.get("use_gpu", int(ray.cluster_resources().get("GPU", 0)) > 0)
    if use_gpu:
        num_workers = int(ray.cluster_resources().get("GPU", 0))
    else:
        num_workers = _num_nodes()

    strategy = kwargs.pop("strategy", get_default_strategy_name())
    backend = get_dist_strategy(strategy).get_ray_trainer_backend(**kwargs)

    # Remove params used by strategy but not the trainer here
    kwargs.pop("nics", None)

    defaults = dict(
        backend=backend,
        strategy=strategy,
        num_workers=num_workers,
        use_gpu=use_gpu,
        resources_per_worker={
            "CPU": 0 if use_gpu else 1,
            "GPU": 1 if use_gpu else 0,
        },
    )
    return {**defaults, **kwargs}


def _create_dask_engine(**kwargs):
    from ludwig.data.dataframe.dask import DaskEngine

    return DaskEngine(**kwargs)


def _create_modin_engine(**kwargs):
    from ludwig.data.dataframe.modin import ModinEngine

    return ModinEngine(**kwargs)


def _create_pandas_engine(**kwargs):
    from ludwig.data.dataframe.pandas import PandasEngine

    return PandasEngine(**kwargs)


_engine_registry = {
    "dask": _create_dask_engine,
    "modin": _create_modin_engine,
    "pandas": _create_pandas_engine,
}


def _get_df_engine(processor):
    logger.info(f"Ray processor params: {processor}")
    if processor is None:
        # TODO ray: find an informed way to set the parallelism, in practice
        #  it looks like Dask handles this well on its own most of the time
        return _create_dask_engine()

    processor_kwargs = processor.copy()

    dtype = processor_kwargs.pop("type", "dask")
    engine_cls = _engine_registry.get(dtype)

    return engine_cls(**processor_kwargs)


def _local_size() -> int:
    return torch.cuda.device_count() if torch.cuda.is_available() else 1


def train_fn(
    executable_kwargs: Dict[str, Any] = None,
    model_ref: ObjectRef = None,  # noqa: F821
    training_set_metadata: TrainingSetMetadataDict = None,
    features: Dict[str, Dict] = None,
    **kwargs,
):
    # Pin GPU before loading the model to prevent memory leaking onto other devices
    initialize_pytorch(local_rank=session.get_local_rank(), local_size=_local_size())
    distributed = get_current_dist_strategy(allow_local=False)()
    try:
        train_shard = RayDatasetShard(
            session.get_dataset_shard("train"),
            features,
            training_set_metadata,
        )

        try:
            val_shard = session.get_dataset_shard("val")
        except KeyError:
            val_shard = None

        if val_shard is not None:
            val_shard = RayDatasetShard(val_shard, features, training_set_metadata)

        try:
            test_shard = session.get_dataset_shard("test")
        except KeyError:
            test_shard = None

        if test_shard is not None:
            test_shard = RayDatasetShard(test_shard, features, training_set_metadata)

        model = ray.get(model_ref)
        device = get_torch_device()
        model = model.to(device)

        trainer = RemoteTrainer(model=model, distributed=distributed, report_tqdm_to_ray=True, **executable_kwargs)
        results = trainer.train(train_shard, val_shard, test_shard, **kwargs)

        if results is not None:
            # only return the model state dict back to the head node.
            trained_model, *args = results
            results = (trained_model.cpu().state_dict(), *args)

        torch.cuda.empty_cache()

        # Passing objects containing Torch tensors as metrics is not supported as it will throw an
        # exception on deserialization, so create a checkpoint and return via session.report() along
        # with the path of the checkpoint
        ckpt = Checkpoint.from_dict({"state_dict": results})
        torch_ckpt = TorchCheckpoint.from_checkpoint(ckpt)

        # The checkpoint is put in the object store and then retrieved by the Trainable actor to be reported to Tune.
        # It is also persisted on disk by the Trainable (and synced to cloud, if configured to do so)
        # The result object returned from trainer.fit() contains the metrics from the last session.report() call.
        # So, make a final call to session.report with the train_results object above.
        session.report(
            metrics={
                "validation_field": trainer.validation_field,
                "validation_metric": trainer.validation_metric,
            },
            checkpoint=torch_ckpt,
        )

    except Exception:
        logger.exception("Exception raised during training by one of the workers")
        raise

    finally:
        distributed.shutdown()


@ray.remote(max_calls=1)
def tune_batch_size_fn(
    dataset: RayDataset = None,
    data_loader_kwargs: Dict[str, Any] = None,
    executable_kwargs: Dict[str, Any] = None,
    model: ECD = None,  # noqa: F821
    ludwig_config: ModelConfigDict = None,
    training_set_metadata: TrainingSetMetadataDict = None,
    features: Dict[str, Dict] = None,
    **kwargs,
) -> int:
    # Pin GPU before loading the model to prevent memory leaking onto other devices
    #
    # As of Ray >= 2.1, to use ray.air.session.get_local_rank(), you need to be inside a train session
    # or a tune session. In Ludwig's current code implementation, batch size tuning doesn't get instantiated
    # inside of a RayTrainer class, so we manually set the local_rank to 0 so that it picks up the right
    # device to tune batch size on.
    initialize_pytorch(local_rank=0, local_size=_local_size())
    distributed = get_current_dist_strategy(allow_local=True)()
    try:
        train_shard = RayDatasetShard(
            dataset.ds,
            features,
            training_set_metadata,
        )

        device = get_torch_device()
        model = model.to(device)

        trainer = RemoteTrainer(model=model, distributed=distributed, **executable_kwargs)
        return trainer.tune_batch_size(ludwig_config, train_shard, **kwargs)
    finally:
        torch.cuda.empty_cache()
        distributed.shutdown()


@DeveloperAPI
class TqdmCallback(ray.tune.callback.Callback):
    """Class for a custom Ray callback that updates tqdm progress bars in the driver process."""

    def __init__(self) -> None:
        """Constructor for TqdmCallback."""
        super().__init__()
        self.progress_bars = {}

    def on_trial_result(self, iteration, trials, trial, result, **info):
        """Called after receiving a result from a trial
        https://docs.ray.io/en/latest/_modules/ray/tune/callback.html#Callback.on_trial_result."""
        progress_bar_opts = result.get("progress_bar")
        if not progress_bar_opts:
            return
        # Skip commands received by non-coordinators
        if not progress_bar_opts["is_coordinator"]:
            return
        _id = progress_bar_opts["id"]
        action = progress_bar_opts.pop("action")
        if action == "create":
            progress_bar_config = progress_bar_opts.get("config")
            self.progress_bars[_id] = tqdm.tqdm(**progress_bar_config)
        elif action == "close":
            self.progress_bars[_id].close()
        elif action == "update":
            update_by = progress_bar_opts.pop("update_by")
            self.progress_bars[_id].update(update_by)


@contextlib.contextmanager
def create_runner(**kwargs):
    trainer_kwargs = get_trainer_kwargs(**kwargs)
    yield RayAirRunner(trainer_kwargs)


class RayAirRunner:
    def __init__(self, trainer_kwargs: Dict[str, Any]) -> None:
        trainer_kwargs = copy.copy(trainer_kwargs)
        self.backend_config = trainer_kwargs.pop("backend", None)
        self.strategy = trainer_kwargs.pop("strategy", get_default_strategy_name())

        if "max_retries" in trainer_kwargs:
            logger.warning("`max_retries` is no longer supported as a trainer argument in Ray backend. Ignoring it.")
            del trainer_kwargs["max_retries"]

        # When training on GPU, you want to pack workers together to limit network latency during
        # allreduce. Conversely, for CPU training you want to spread out the workers to limit
        # CPU and memory contention and avoid too many workers on a single machine.
        strategy = "PACK" if trainer_kwargs.get("use_gpu", False) else "SPREAD"
        # Ray Tune automatically creates a PlacementGroupFactory from the ScalingConfig internally
        self.scaling_config = ScalingConfig(
            placement_strategy=strategy,
            # Override the default of 1 to prevent unnecessary CPU usage.
            trainer_resources={"CPU": 0},
            **trainer_kwargs,
        )

    def _get_dataset_configs(
        self,
        datasets: Dict[str, Any],
        stream_window_size: Dict[str, Union[None, float]],
        data_loader_kwargs: Dict[str, Any],
    ) -> Dict[str, DatasetConfig]:
        """Generates DatasetConfigs for each dataset passed into the trainer."""
        dataset_configs = {}
        for dataset_name, _ in datasets.items():
            if _ray230:
                # DatasetConfig.use_stream_api and DatasetConfig.stream_window_size have been removed as of Ray 2.3.
                # We need to use DatasetConfig.max_object_store_memory_fraction instead -> default to 20% when windowing
                # is enabled unless the end user specifies a different fraction.
                # https://docs.ray.io/en/master/ray-air/check-ingest.html?highlight=max_object_store_memory_fraction#enabling-streaming-ingest # noqa
                dataset_conf = DatasetConfig(
                    split=True,
                    max_object_store_memory_fraction=stream_window_size.get(dataset_name),
                )
            else:
                dataset_conf = DatasetConfig(
                    split=True,
                    use_stream_api=True,
                    stream_window_size=stream_window_size.get(dataset_name),
                )

            if dataset_name == "train":
                # Mark train dataset as always required
                dataset_conf.required = True
                # Check data loader kwargs to see if shuffle should be enabled for the
                # train dataset. global_shuffle is False by default for all other datasets.
                dataset_conf.global_shuffle = data_loader_kwargs.get("shuffle", True)
            dataset_configs[dataset_name] = dataset_conf
        return dataset_configs

    def run(
        self,
        train_loop_per_worker: Callable,
        config: Dict[str, Any],
        dataset: Dict[str, Any],
        data_loader_kwargs: Dict[str, Any],
        stream_window_size: Dict[str, Union[None, float]],
        callbacks: List[Any] = [],
    ) -> Tuple[Dict, TorchCheckpoint]:
        trainer_cls, kwargs = get_dist_strategy(self.strategy).get_trainer_cls(self.backend_config)
        trainer = trainer_cls(
            train_loop_per_worker=train_loop_per_worker,
            train_loop_config=config,
            datasets=dataset,
            scaling_config=self.scaling_config,
            dataset_config=self._get_dataset_configs(dataset, stream_window_size, data_loader_kwargs),
            run_config=RunConfig(callbacks=callbacks, verbose=0),
            **kwargs,
        )
        return trainer.fit()


@register_ray_trainer(MODEL_ECD, default=True)
class RayTrainerV2(BaseTrainer):
    def __init__(
        self,
        model: BaseModel,
        trainer_kwargs: Dict[str, Any],
        data_loader_kwargs: Dict[str, Any],
        executable_kwargs: Dict[str, Any],
        **kwargs,
    ):
        self.model = model.cpu()
        self.data_loader_kwargs = data_loader_kwargs or {}
        self.executable_kwargs = executable_kwargs
        self.trainer_kwargs = trainer_kwargs
        self._validation_field = None
        self._validation_metric = None

    @staticmethod
    def get_schema_cls():
        return ECDTrainerConfig

    def train(
        self,
        training_set: RayDataset,
        validation_set: Optional[RayDataset] = None,
        test_set: Optional[RayDataset] = None,
        **kwargs,
    ):
        executable_kwargs = self.executable_kwargs
        kwargs = {
            "training_set_metadata": training_set.training_set_metadata,
            "features": training_set.features,
            **kwargs,
        }

        dataset = {"train": training_set.ds}
        stream_window_size = {"train": training_set.window_size_bytes}
        if validation_set is not None:
            dataset["val"] = validation_set.ds
            stream_window_size["val"] = validation_set.window_size_bytes
        if test_set is not None:
            dataset["test"] = test_set.ds
            stream_window_size["test"] = test_set.window_size_bytes

        with create_runner(**self.trainer_kwargs) as runner:
            trainer_results = runner.run(
                lambda config: train_fn(**config),
                config={
                    "executable_kwargs": executable_kwargs,
                    "model_ref": ray.put(self.model),
                    **kwargs,
                },
                callbacks=[TqdmCallback()],
                data_loader_kwargs=self.data_loader_kwargs,
                dataset=dataset,
                stream_window_size=stream_window_size,
            )

        # Set validation field and metric used by trainer
        self._validation_field = trainer_results.metrics["validation_field"]
        self._validation_metric = trainer_results.metrics["validation_metric"]

        # Load model from checkpoint
        ckpt = TorchCheckpoint.from_checkpoint(trainer_results.checkpoint)
        results = ckpt.to_dict()["state_dict"]

        # load state dict back into the model
        state_dict, *args = results
        self.model.load_state_dict(state_dict)
        results = (self.model, *args)

        return results

    def train_online(self, *args, **kwargs):
        # TODO: When this is implemented we also need to update the
        # Tqdm flow to report back the callback
        raise NotImplementedError()

    def tune_batch_size(
        self,
        config: ModelConfigDict,
        training_set: RayDataset,
        **kwargs,
    ) -> int:
        return ray.get(
            tune_batch_size_fn.options(num_cpus=self.num_cpus, num_gpus=self.num_gpus).remote(
                dataset=training_set,
                data_loader_kwargs=self.data_loader_kwargs,
                executable_kwargs=self.executable_kwargs,
                model=ray.put(self.model),
                ludwig_config=config,
                training_set_metadata=training_set.training_set_metadata,
                features=training_set.features,
                **kwargs,
            )
        )

    @property
    def validation_field(self):
        return self._validation_field

    @property
    def validation_metric(self):
        return self._validation_metric

    @property
    def config(self) -> ECDTrainerConfig:
        return self.executable_kwargs["config"]

    @property
    def batch_size(self) -> int:
        return self.config.batch_size

    @batch_size.setter
    def batch_size(self, value: int):
        self.config.batch_size = value

    @property
    def eval_batch_size(self) -> int:
        return self.config.eval_batch_size if self.config.eval_batch_size is not None else self.config.batch_size

    @eval_batch_size.setter
    def eval_batch_size(self, value: int):
        self.config.eval_batch_size = value

    @property
    def resources_per_worker(self) -> Dict[str, Any]:
        trainer_kwargs = get_trainer_kwargs(**self.trainer_kwargs)
        return trainer_kwargs.get("resources_per_worker", {})

    @property
    def num_cpus(self) -> int:
        return self.resources_per_worker.get("CPU", 1)

    @property
    def num_gpus(self) -> int:
        return self.resources_per_worker.get("GPU", 0)

    def shutdown(self):
        pass


class HorovodRemoteTrainer(RemoteTrainer):
    def __init__(self, **kwargs):
        distributed = get_current_dist_strategy(allow_local=False)()
        super().__init__(distributed=distributed, **kwargs)


def eval_fn(
    predictor_kwargs: Dict[str, Any] = None,
    model_ref: ObjectRef = None,  # noqa: F821
    training_set_metadata: TrainingSetMetadataDict = None,
    features: Dict[str, Dict] = None,
    **kwargs,
):
    # Pin GPU before loading the model to prevent memory leaking onto other devices
    initialize_pytorch(local_rank=session.get_local_rank(), local_size=_local_size())
    distributed = get_current_dist_strategy(allow_local=False)()
    try:
        eval_shard = RayDatasetShard(
            session.get_dataset_shard("eval"),
            features,
            training_set_metadata,
        )

        model = ray.get(model_ref)
        device = get_torch_device()
        model = model.to(device)

        predictor = RemotePredictor(model=model, distributed=distributed, report_tqdm_to_ray=True, **predictor_kwargs)
        results = predictor.batch_evaluation(eval_shard, **kwargs)

        # The result object returned from trainer.fit() contains the metrics from the last session.report() call.
        # So, make a final call to session.report with the eval_results object above.
        session.report(metrics={"eval_results": results})
    finally:
        torch.cuda.empty_cache()
        distributed.shutdown()


class RayPredictor(BasePredictor):
    def __init__(
        self, model: BaseModel, df_engine: DataFrameEngine, trainer_kwargs, data_loader_kwargs, **predictor_kwargs
    ):
        self.batch_size = predictor_kwargs["batch_size"]
        self.trainer_kwargs = trainer_kwargs
        self.data_loader_kwargs = data_loader_kwargs
        self.predictor_kwargs = predictor_kwargs
        self.actor_handles = []
        self.model = model.cpu()
        self.df_engine = df_engine

    def get_trainer_kwargs(self) -> Dict[str, Any]:
        return get_trainer_kwargs(**self.trainer_kwargs)

    def get_resources_per_worker(self) -> Tuple[int, int]:
        trainer_kwargs = self.get_trainer_kwargs()
        resources_per_worker = trainer_kwargs.get("resources_per_worker", {})
        num_gpus = resources_per_worker.get("GPU", 0)
        num_cpus = resources_per_worker.get("CPU", (1 if num_gpus == 0 else 0))
        return num_cpus, num_gpus

    def batch_predict(self, dataset: RayDataset, *args, collect_logits: bool = False, **kwargs):
        self._check_dataset(dataset)

        predictor_kwargs = self.predictor_kwargs
        output_columns = get_output_columns(self.model.output_features, include_logits=collect_logits)
        batch_predictor = self.get_batch_infer_model(
            self.model,
            predictor_kwargs,
            output_columns,
            dataset.features,
            dataset.training_set_metadata,
            *args,
            collect_logits=collect_logits,
            **kwargs,
        )

        num_cpus, num_gpus = self.get_resources_per_worker()

        with tensor_extension_casting(False):
            predictions = dataset.ds.map_batches(
                batch_predictor,
                batch_size=self.batch_size,
                compute="actors",
                batch_format="pandas",
                num_cpus=num_cpus,
                num_gpus=num_gpus,
            )
            predictions = self.df_engine.from_ray_dataset(predictions)

        for of_feature in self.model.output_features.values():
            predictions = of_feature.unflatten(predictions)

        return predictions

    def predict_single(self, batch):
        raise NotImplementedError("predict_single can only be called on a local predictor")

    def batch_evaluation(
        self,
        dataset: RayDataset,
        collect_predictions: bool = False,
        collect_logits=False,
        **kwargs,
    ):
        # We need to be in a Horovod context to collect the aggregated metrics, since it relies on collective
        # communication ops. However, Horovod is not suitable for transforming one big dataset to another. For that
        # we will use Ray Datasets. Therefore, we break this up into two separate steps, and two passes over the
        # dataset. In the future, we can explore ways to combine these into a single step to reduce IO.
        with create_runner(**self.trainer_kwargs) as runner:
            # Collect eval metrics by distributing work across nodes / gpus with Horovod
            datasets = {"eval": dataset.ds}
            stream_window_size = {"eval": dataset.window_size_bytes}
            predictor_kwargs = {**self.predictor_kwargs, "collect_predictions": False}
            eval_results = runner.run(
                lambda config: eval_fn(**config),
                config={
                    "predictor_kwargs": predictor_kwargs,
                    "model_ref": ray.put(self.model),
                    "training_set_metadata": dataset.training_set_metadata,
                    "features": dataset.features,
                    **kwargs,
                },
                dataset=datasets,
                data_loader_kwargs=self.data_loader_kwargs,
                stream_window_size=stream_window_size,
            )

        eval_stats = eval_results.metrics["eval_results"][0]

        predictions = None
        if collect_predictions:
            # Collect eval predictions by using Ray Datasets to transform partitions of the data in parallel
            predictions = self.batch_predict(dataset, collect_logits=collect_logits)

        return eval_stats, predictions

    def batch_collect_activations(self, model, *args, **kwargs):
        raise NotImplementedError("Ray backend does not support collecting activations at this time.")

    def _check_dataset(self, dataset):
        if not isinstance(dataset, RayDataset):
            raise RuntimeError(f"Ray backend requires RayDataset for inference, " f"found: {type(dataset)}")

    def shutdown(self):
        for handle in self.actor_handles:
            ray.kill(handle)
        self.actor_handles.clear()

    def get_batch_infer_model(
        self,
        model: "LudwigModel",  # noqa: F821
        predictor_kwargs: Dict[str, Any],
        output_columns: List[str],
        features: Dict[str, Dict],
        training_set_metadata: TrainingSetMetadataDict,
        *args,
        **kwargs,
    ):
        model_ref = ray.put(model)

        class BatchInferModel:
            def __init__(self):
                model = ray.get(model_ref)
                device = get_torch_device()
                self.model = model.to(device)

                self.output_columns = output_columns
                self.features = features
                self.training_set_metadata = training_set_metadata
                self.reshape_map = {
                    f[PROC_COLUMN]: training_set_metadata[f[NAME]].get("reshape") for f in features.values()
                }
                predictor = Predictor(model, **predictor_kwargs)
                self.predict = partial(predictor.predict_single, *args, **kwargs)

            def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
                dataset = self._prepare_batch(df)
                predictions = self.predict(batch=dataset).set_index(df.index)
                for output_feature in self.model.output_features.values():
                    predictions = output_feature.flatten(predictions)
                ordered_predictions = predictions[self.output_columns]
                return ordered_predictions

            def _prepare_batch(self, batch: pd.DataFrame) -> Dict[str, np.ndarray]:
                res = {}
                for c in self.features.keys():
                    if self.features[c][TYPE] not in _SCALAR_TYPES:
                        # Ensure columns stacked instead of turned into np.array([np.array, ...], dtype=object) objects
                        res[c] = np.stack(batch[c].values)
                    else:
                        res[c] = batch[c].to_numpy()

                for c in self.features.keys():
                    reshape = self.reshape_map.get(c)
                    if reshape is not None:
                        res[c] = res[c].reshape((-1, *reshape))

                return res

        return BatchInferModel


class RayBackend(RemoteTrainingMixin, Backend):
    BACKEND_TYPE = "ray"

    def __init__(self, processor=None, trainer=None, loader=None, preprocessor_kwargs=None, **kwargs):
        super().__init__(dataset_manager=RayDatasetManager(self), **kwargs)
        self._preprocessor_kwargs = preprocessor_kwargs or {}
        self._df_engine = _get_df_engine(processor)
        self._horovod_kwargs = trainer or {}
        self._pytorch_kwargs = {}
        self._data_loader_kwargs = loader or {}
        self._preprocessor_pg = None

    def initialize(self):
        initialize_ray()

        dask.config.set(scheduler=ray_dask_get)
        # Disable placement groups on dask
        dask.config.set(annotations={"ray_remote_args": {"placement_group": None}})

    def generate_bundles(self, num_cpu):
        # Ray requires that each bundle be scheduleable on a single node.
        # So a bundle of 320 cpus would never get scheduled. For now a simple heuristic
        # to be used is to just request 1 cpu at a time.
        return [{"CPU": 1} for _ in range(int(num_cpu))]

    @contextlib.contextmanager
    def provision_preprocessing_workers(self):
        num_cpu = self._preprocessor_kwargs.get("num_cpu")
        if not num_cpu:
            logger.info(
                "Backend config has num_cpu not set." " provision_preprocessing_workers() is a no-op in this case."
            )
            yield
        else:
            bundles = self.generate_bundles(num_cpu)
            logger.info("Requesting bundles of %s for preprocessing", bundles)
            self._preprocessor_pg = placement_group(bundles)
            ready = self._preprocessor_pg.wait(FIFTEEN_MINS_IN_S)

            if not ready:
                remove_placement_group(self._preprocessor_pg)
                raise TimeoutError(
                    "Ray timed out in provisioning the placement group for preprocessing."
                    f" {num_cpu} CPUs were requested but were unable to be provisioned."
                )

            logger.info("%s CPUs were requested and successfully provisioned", num_cpu)
            try:
                with dask.config.set(annotations={"ray_remote_args": {"placement_group": self._preprocessor_pg}}):
                    yield
            finally:
                self._release_preprocessing_workers()

    def _release_preprocessing_workers(self):
        if self._preprocessor_pg is not None:
            remove_placement_group(self._preprocessor_pg)
        self._preprocessor_pg = None

    def initialize_pytorch(self, **kwargs):
        # Make sure we don't claim any GPU resources on the head node
        initialize_pytorch(gpus=-1)
        self._pytorch_kwargs = kwargs

    def create_trainer(self, model: BaseModel, **kwargs) -> "BaseTrainer":  # noqa: F821
        executable_kwargs = {**kwargs, **self._pytorch_kwargs}

        trainer_cls = get_from_registry(model.type(), ray_trainers_registry)

        all_kwargs = {
            "model": model,
            "trainer_kwargs": self._horovod_kwargs,
            "data_loader_kwargs": self._data_loader_kwargs,
            "executable_kwargs": executable_kwargs,
        }
        all_kwargs.update(kwargs)

        return trainer_cls(**all_kwargs)

    def create_predictor(self, model: BaseModel, **kwargs):
        executable_kwargs = {**kwargs, **self._pytorch_kwargs}
        return RayPredictor(
            model,
            self.df_engine,
            self._horovod_kwargs,
            self._data_loader_kwargs,
            **executable_kwargs,
        )

    @property
    def distributed_kwargs(self):
        return self._horovod_kwargs

    @distributed_kwargs.setter
    def distributed_kwargs(self, value):
        self._horovod_kwargs = value

    @property
    def df_engine(self):
        return self._df_engine

    @property
    def supports_multiprocessing(self):
        return False

    def read_binary_files(
        self, column: Series, map_fn: Optional[Callable] = None, file_size: Optional[int] = None
    ) -> Series:
        column = column.fillna(np.nan).replace([np.nan], [None])  # normalize NaNs to None

        # Assume that the list of filenames is small enough to fit in memory. Should be true unless there
        # are literally billions of filenames.
        # TODO(travis): determine if there is a performance penalty to passing in individual files instead of
        #  a directory. If so, we can do some preprocessing to determine if it makes sense to read the full directory
        #  then filter out files as a postprocessing step (depending on the ratio of included to excluded files in
        #  the directory). Based on a preliminary look at how Ray handles directory expansion to files, it looks like
        #  there should not be any difference between providing a directory versus a list of files.
        pd_column = self.df_engine.compute(column)
        fnames = pd_column.values.tolist()
        idxs = pd_column.index.tolist()

        # Sample a filename to extract the filesystem info
        sample_fname = fnames[0]
        if isinstance(sample_fname, str):
            fs, _ = get_fs_and_path(sample_fname)

            read_datasource_fn_kwargs = {
                "path_and_idxs": list(zip(fnames, idxs)),
                "filesystem": PyFileSystem(FSSpecHandler(fs)),
            }
            if self.df_engine.partitioned and file_size is not None:
                # Heuristic to determine parallelism: if the average file size is known (in bytes), then we can
                # extrapolate to determine the total file size. We aim to have ~50MB partitions (5e7 bytes), so we
                # set parallelism to be the total size / 50MB.
                total_size = file_size * len(fnames)
                parallelism = int(total_size / 5e7)
                # Only set parallelism if it matches or exceeds the Ray default kwarg for parallelism
                read_datasource_fn_kwargs["parallelism"] = max(RAY_DEFAULT_PARALLELISM, parallelism)

            # The resulting column is named "value", which is a dict with two keys: "idx" and "data".
            ds = ray.data.read_datasource(BinaryIgnoreNoneTypeDatasource(), **read_datasource_fn_kwargs)
            df = self.df_engine.from_ray_dataset(ds)
            # Persist the dataframe to prevent re-reading binary files on each subsequent map_objects call
            df = self.df_engine.persist(df)
            df["idx"] = self.df_engine.map_objects(df["value"], lambda row: int(row["idx"]))
            df["value"] = self.df_engine.map_objects(df["value"], lambda row: row["data"])
            df = df.rename(columns={"value": column.name})
        else:
            # Assume the path has already been read in, so just convert directly to a dataset
            # Name the column "value" to match the behavior of the above
            df = column.to_frame(name=column.name)
            df["idx"] = df.index

        if map_fn is not None:
            df[column.name] = self.df_engine.map_objects(df[column.name], map_fn)

        if "idx" in df.columns:
            df = df.set_index("idx", drop=True)
            df = self.df_engine.map_partitions(
                df, lambda pd_df: set_index_name(pd_df, column.index.name), meta={column.name: "object"}
            )
        return df[column.name]

    @property
    def num_nodes(self) -> int:
        if not ray.is_initialized():
            return 1
        return len(ray.nodes())

    def get_available_resources(self) -> Resources:
        resources = ray.cluster_resources()
        return Resources(cpus=resources.get("CPU", 0), gpus=resources.get("GPU", 0))

    def max_concurrent_trials(self, hyperopt_config: HyperoptConfigDict) -> Union[int, None]:
        cpus_per_trial = hyperopt_config[EXECUTOR].get(CPU_RESOURCES_PER_TRIAL, 1)
        num_cpus_available = self.get_available_resources().cpus

        # No actors will compete for ray datasets tasks dataset tasks are cpu bound
        if cpus_per_trial == 0:
            return None

        if num_cpus_available < 2:
            logger.warning(
                "At least 2 CPUs are required for hyperopt when using a RayBackend, but only found "
                f"{num_cpus_available}. If you are not using an auto-scaling Ray cluster, your hyperopt "
                "trials may hang."
            )

        # Ray requires at least 1 free CPU to ensure trials don't stall
        max_possible_trials = int(num_cpus_available // cpus_per_trial) - 1

        # Users may be using an autoscaling cluster, so return None
        if max_possible_trials < 1:
            logger.warning(
                f"Hyperopt trials will request {cpus_per_trial} CPUs in addition to CPUs needed for Ray Datasets, "
                f" but only {num_cpus_available} CPUs are currently available. If you are not using an auto-scaling "
                " Ray cluster, your hyperopt trials may hang."
            )
            return None

        return max_possible_trials

    def tune_batch_size(self, evaluator_cls: Type[BatchSizeEvaluator], dataset_len: int) -> int:
        return ray.get(
            _tune_batch_size_fn.options(**self._get_transform_kwargs()).remote(
                evaluator_cls,
                dataset_len,
            )
        )

    def batch_transform(self, df: DataFrame, batch_size: int, transform_fn: Callable) -> DataFrame:
        ds = self.df_engine.to_ray_dataset(df)
        ds = ds.map_batches(
            transform_fn, batch_size=batch_size, compute="actors", batch_format="pandas", **self._get_transform_kwargs()
        )
        return self.df_engine.from_ray_dataset(ds)

    def _get_transform_kwargs(self) -> Dict[str, Any]:
        trainer_kwargs = get_trainer_kwargs(**self._horovod_kwargs)
        resources_per_worker = trainer_kwargs.get("resources_per_worker", {})
        num_gpus = resources_per_worker.get("GPU", 0)
        num_cpus = resources_per_worker.get("CPU", (1 if num_gpus == 0 else 0))
        return dict(num_cpus=num_cpus, num_gpus=num_gpus)


@ray.remote(max_calls=1)
def _tune_batch_size_fn(evaluator_cls: Type[BatchSizeEvaluator], dataset_len: int) -> int:
    evaluator = evaluator_cls()
    return evaluator.select_best_batch_size(dataset_len)


def initialize_ray():
    if not ray.is_initialized():
        try:
            ray.init("auto", ignore_reinit_error=True)
        except ConnectionError:
            init_ray_local()


def init_ray_local():
    logger.info("Initializing new Ray cluster...")
    ray.init(ignore_reinit_error=True)



#! /usr/bin/env python
# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np

from ludwig.backend import Backend
from ludwig.models.ecd import ECD


class Calibrator:
    """Calibrator calibrates the output probabilities of a model."""

    def __init__(self, model: ECD, backend: Backend, batch_size: int = 128):
        self.model = model
        self.backend = backend
        self.batch_size = batch_size

    def calibration_enabled(self):
        """Calibration is enabled if the config requests calibration for any output feature.

        If no output features have calibration enabled, the calibration phase should be skipped.
        """
        return any(o.calibration_module is not None for o in self.model.output_features.values())

    def train_calibration(self, dataset, dataset_name: str):
        """Calibrates model output probabilities on validation set after training.

        This works well for most datasets, though it may fail for some difficult or extremely imbalanced datasets.
        """
        if not self.calibration_enabled():
            # Early out if no output features have calibration enabled.
            return
        with self.backend.create_predictor(self.model, batch_size=self.batch_size) as predictor:
            metrics, predictions = predictor.batch_evaluation(
                dataset, collect_predictions=True, collect_logits=True, dataset_name=dataset_name
            )

        dataset_df = dataset.to_df()
        for output_feature in self.model.output_features.values():
            if output_feature.calibration_module is not None:
                feature_logits_key = f"{output_feature.feature_name}_logits"
                if feature_logits_key in predictions:
                    feature_logits = self.backend.df_engine.compute(predictions[feature_logits_key])
                    feature_labels = self.backend.df_engine.compute(dataset_df[output_feature.proc_column])
                    output_feature.calibration_module.train_calibration(
                        np.stack(feature_logits.values, axis=0), np.stack(feature_labels.values, axis=0)
                    )



import contextlib
import os
from typing import List
from unittest.mock import patch

import pytest
import torch

from ludwig.utils.torch_utils import (
    _get_torch_init_params,
    _set_torch_init_params,
    initialize_pytorch,
    sequence_length_2D,
    sequence_length_3D,
)


@pytest.mark.parametrize("input_sequence", [[[0, 1, 1], [2, 0, 0], [3, 3, 3]]])
@pytest.mark.parametrize("expected_output", [[3, 2, 3]])
def test_sequence_length_2D(input_sequence: List[List[int]], expected_output: List[int]):
    output_seq_length = sequence_length_2D(torch.tensor(input_sequence))
    assert torch.equal(torch.tensor(expected_output), output_seq_length)


@pytest.mark.parametrize("input_sequence", [[[[-1, 0, 1], [1, -2, 0]], [[0, 0, 0], [3, 0, -2]]]])
@pytest.mark.parametrize("expected_output", [[2, 1]])
def test_sequence_length_3D(input_sequence: List[List[List[int]]], expected_output: List[int]):
    input_sequence = torch.tensor(input_sequence, dtype=torch.int32)
    expected_output = torch.tensor(expected_output, dtype=torch.int32)
    output_seq_length = sequence_length_3D(input_sequence)
    assert torch.equal(expected_output, output_seq_length)


@contextlib.contextmanager
def clean_params():
    prev = _get_torch_init_params()
    try:
        _set_torch_init_params(None)
        if "CUDA_VISIBLE_DEVICES" in os.environ:
            del os.environ["CUDA_VISIBLE_DEVICES"]
        yield
    finally:
        _set_torch_init_params(prev)


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_only_once(mock_torch):
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4
    with clean_params():
        # During first time initialization, set pytorch parallelism
        initialize_pytorch(allow_parallel_threads=False)
        mock_torch.set_num_threads.assert_called_once()
        mock_torch.set_num_interop_threads.assert_called_once()

        # Reset call counts on all threading calls
        mock_torch.reset_mock()

        # In the second call to initialization, avoid calling these methods again, as pytorch
        # will raise an exception
        initialize_pytorch(allow_parallel_threads=False)
        mock_torch.set_num_threads.assert_not_called()
        mock_torch.set_num_interop_threads.assert_not_called()

    # No GPUs were specified, so this should not have been called even once
    mock_torch.cuda.memory.set_per_process_memory_fraction.assert_not_called()


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_with_gpu_list(mock_torch):
    # For test purposes, these devices can be anything, we just need to be able to uniquely
    # identify them.
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4
    with clean_params():
        initialize_pytorch(gpus=[1, 2])
    assert os.environ["CUDA_VISIBLE_DEVICES"] == "1,2"


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_with_gpu_string(mock_torch):
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4
    with clean_params():
        initialize_pytorch(gpus="1,2")
    assert os.environ["CUDA_VISIBLE_DEVICES"] == "1,2"


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_with_gpu_int(mock_torch):
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4
    with clean_params():
        initialize_pytorch(gpus=1)
    mock_torch.cuda.set_device.assert_called_with(1)
    assert "CUDA_VISIBLE_DEVICES" not in os.environ


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_without_gpu(mock_torch):
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4
    with clean_params():
        initialize_pytorch(gpus=-1)
    assert os.environ["CUDA_VISIBLE_DEVICES"] == ""


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_with_horovod(mock_torch):
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4

    with clean_params():
        initialize_pytorch(local_rank=1, local_size=4)

    mock_torch.cuda.set_device.assert_called_with(1)
    assert "CUDA_VISIBLE_DEVICES" not in os.environ


@patch("ludwig.utils.torch_utils.warnings")
@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_with_horovod_bad_local_rank(mock_torch, mock_warnings):
    """In this scenario, the local_size 5 is out of the bounds of the GPU indices."""
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4

    with clean_params():
        initialize_pytorch(local_rank=1, local_size=5)

    assert os.environ["CUDA_VISIBLE_DEVICES"] == ""
    mock_warnings.warn.assert_called()


@patch("ludwig.utils.torch_utils.torch")
def test_initialize_pytorch_with_horovod_explicit_gpus(mock_torch):
    mock_torch.cuda.is_available.return_value = True
    mock_torch.cuda.device_count.return_value = 4

    with clean_params():
        initialize_pytorch(gpus="-1", local_rank=1, local_size=4)

    assert os.environ["CUDA_VISIBLE_DEVICES"] == ""



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import logging
import os
import sys
from typing import List, Optional, Union

import numpy as np
import torchinfo

from ludwig.api import LudwigModel
from ludwig.backend import ALL_BACKENDS, Backend
from ludwig.callbacks import Callback
from ludwig.constants import FULL, TEST, TRAINING, VALIDATION
from ludwig.contrib import add_contrib_callback_args
from ludwig.globals import LUDWIG_VERSION
from ludwig.utils.print_utils import get_logging_level_registry, print_boxed, print_ludwig
from ludwig.utils.strings_utils import make_safe_filename

logger = logging.getLogger(__name__)


def collect_activations(
    model_path: str,
    layers: List[str],
    dataset: str,
    data_format: str = None,
    split: str = FULL,
    batch_size: int = 128,
    output_directory: str = "results",
    gpus: List[str] = None,
    gpu_memory_limit: Optional[float] = None,
    allow_parallel_threads: bool = True,
    callbacks: List[Callback] = None,
    backend: Union[Backend, str] = None,
    **kwargs,
) -> List[str]:
    """Uses the pretrained model to collect the tensors corresponding to a datapoint in the dataset. Saves the
    tensors to the experiment directory.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param layers: (List[str]) list of strings for layer names in the model
        to collect activations.
    :param dataset: (str) source
        containing the data to make predictions.
    :param data_format: (str, default: `None`) format to interpret data
        sources. Will be inferred automatically if not specified.  Valid
        formats are `'auto'`, `'csv'`, `'excel'`, `'feather'`,
        `'fwf'`, `'hdf5'` (cache file produced during previous training),
        `'html'` (file containing a single HTML `<table>`), `'json'`, `'jsonl'`,
        `'parquet'`, `'pickle'` (pickled Pandas DataFrame), `'sas'`, `'spss'`,
        `'stata'`, `'tsv'`.
    :param split: (str, default: `full`) split on which
        to perform predictions. Valid values are `'training'`, `'validation'`,
        `'test'` and `'full'`.
    :param batch_size: (int, default `128`) size of batches for processing.
    :param output_directory: (str, default: `'results'`) the directory that
        will contain the training statistics, TensorBoard logs, the saved
        model and the training progress files.
    :param gpus: (list, default: `None`) list of GPUs that are available
        for training.
    :param gpu_memory_limit: (float: default: `None`) maximum memory fraction
        [0, 1] allowed to allocate per GPU device.
    :param allow_parallel_threads: (bool, default: `True`) allow PyTorch
        to use multithreading parallelism to improve performance at
        the cost of determinism.
    :param callbacks: (list, default: `None`) a list of
        `ludwig.callbacks.Callback` objects that provide hooks into the
        Ludwig pipeline.
    :param backend: (Union[Backend, str]) `Backend` or string name
        of backend to use to execute preprocessing / training steps.

    # Return

    :return: (List[str]) list of filepath to `*.npy` files containing
        the activations.
    """
    logger.info(f"Dataset path: {dataset}")
    logger.info(f"Model path: {model_path}")
    logger.info(f"Output path: {output_directory}")
    logger.info("\n")

    model = LudwigModel.load(
        model_path,
        gpus=gpus,
        gpu_memory_limit=gpu_memory_limit,
        allow_parallel_threads=allow_parallel_threads,
        callbacks=callbacks,
        backend=backend,
    )

    # collect activations
    print_boxed("COLLECT ACTIVATIONS")
    collected_tensors = model.collect_activations(
        layers, dataset, data_format=data_format, split=split, batch_size=batch_size
    )

    # saving
    os.makedirs(output_directory, exist_ok=True)
    saved_filenames = save_tensors(collected_tensors, output_directory)

    logger.info(f"Saved to: {output_directory}")
    return saved_filenames


def collect_weights(model_path: str, tensors: List[str], output_directory: str = "results", **kwargs) -> List[str]:
    """Loads a pretrained model and collects weights.

    # Inputs
    :param model_path: (str) filepath to pre-trained model.
    :param tensors: (list, default: `None`) List of tensor names to collect
        weights
    :param output_directory: (str, default: `'results'`) the directory where
        collected weights will be stored.

    # Return

    :return: (List[str]) list of filepath to `*.npy` files containing
        the weights.
    """
    logger.info(f"Model path: {model_path}")
    logger.info(f"Output path: {output_directory}")
    logger.info("\n")

    model = LudwigModel.load(model_path)

    # collect weights
    print_boxed("COLLECT WEIGHTS")
    collected_tensors = model.collect_weights(tensors)

    # saving
    os.makedirs(output_directory, exist_ok=True)
    saved_filenames = save_tensors(collected_tensors, output_directory)

    logger.info(f"Saved to: {output_directory}")
    return saved_filenames


def save_tensors(collected_tensors, output_directory):
    filenames = []
    for tensor_name, tensor_value in collected_tensors:
        np_filename = os.path.join(output_directory, make_safe_filename(tensor_name) + ".npy")
        np.save(np_filename, tensor_value.detach().cpu().numpy())
        filenames.append(np_filename)
    return filenames


def print_model_summary(model_path: str, **kwargs) -> None:
    """Loads a pretrained model and prints names of weights and layers activations.

    # Inputs
    :param model_path: (str) filepath to pre-trained model.

    # Return
    :return: (`None`)
    """
    model = LudwigModel.load(model_path)
    # Model's dict inputs are wrapped in a list, required by torchinfo.
    logger.info(torchinfo.summary(model.model, input_data=[model.model.get_model_inputs()], depth=20))

    logger.info("\nModules:\n")
    for name, _ in model.model.named_children():
        logger.info(name)

    logger.info("\nParameters:\n")
    for name, _ in model.model.named_parameters():
        logger.info(name)


def cli_collect_activations(sys_argv):
    """Command Line Interface to communicate with the collection of tensors and there are several options that can
    specified when calling this function:

    --data_csv: Filepath for the input csv
    --data_hdf5: Filepath for the input hdf5 file, if there is a csv file, this
                 is not read
    --d: Refers to the dataset type of the file being read, by default is
         *generic*
    --s: Refers to the split of the data, can be one of: train, test,
         validation, full
    --m: Input model that is necessary to collect to the tensors, this is a
         required *option*
    --t: Tensors to collect
    --od: Output directory of the model, defaults to results
    --bs: Batch size
    --g: Number of gpus that are to be used
    --gf: Fraction of each GPUs memory to use.
    --v: Verbose: Defines the logging level that the user will be exposed to
    """
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model and uses it collect "
        "tensors for each datapoint in the dataset.",
        prog="ludwig collect_activations",
        usage="%(prog)s [options]",
    )

    # ---------------
    # Data parameters
    # ---------------
    parser.add_argument("--dataset", help="input data file path", required=True)
    parser.add_argument(
        "--data_format",
        help="format of the input data",
        default="auto",
        choices=[
            "auto",
            "csv",
            "excel",
            "feather",
            "fwf",
            "hdf5",
            "html" "tables",
            "json",
            "jsonl",
            "parquet",
            "pickle",
            "sas",
            "spss",
            "stata",
            "tsv",
        ],
    )
    parser.add_argument(
        "-s",
        "--split",
        default=FULL,
        choices=[TRAINING, VALIDATION, TEST, FULL],
        help="the split to obtain the model activations from",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)
    parser.add_argument("-lyr", "--layers", help="tensors to collect", nargs="+", required=True)

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        "-od", "--output_directory", type=str, default="results", help="directory that contains the results"
    )

    # ------------------
    # Generic parameters
    # ------------------
    parser.add_argument("-bs", "--batch_size", type=int, default=128, help="size of batches")

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument("-g", "--gpus", type=int, default=0, help="list of gpu to use")
    parser.add_argument(
        "-gml",
        "--gpu_memory_limit",
        type=float,
        default=None,
        help="maximum memory fraction [0, 1] allowed to allocate per GPU device",
    )
    parser.add_argument(
        "-dpt",
        "--disable_parallel_threads",
        action="store_false",
        dest="allow_parallel_threads",
        help="disable PyTorch from using multithreading for reproducibility",
    )
    parser.add_argument(
        "-b",
        "--backend",
        help="specifies backend to use for parallel / distributed execution, "
        "defaults to local execution or Horovod if called using horovodrun",
        choices=ALL_BACKENDS,
    )
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("collect_activations", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.collect")

    print_ludwig("Collect Activations", LUDWIG_VERSION)

    collect_activations(**vars(args))


def cli_collect_weights(sys_argv):
    """Command Line Interface to collecting the weights for the model.

    --m: Input model that is necessary to collect to the tensors, this is a
         required *option*
    --t: Tensors to collect
    --od: Output directory of the model, defaults to results
    --v: Verbose: Defines the logging level that the user will be exposed to
    """
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model " "and uses it collect weights.",
        prog="ludwig collect_weights",
        usage="%(prog)s [options]",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)
    parser.add_argument("-t", "--tensors", help="tensors to collect", nargs="+", required=True)

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        "-od", "--output_directory", type=str, default="results", help="directory that contains the results"
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("collect_weights", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.collect")

    print_ludwig("Collect Weights", LUDWIG_VERSION)

    collect_weights(**vars(args))


def cli_collect_summary(sys_argv):
    """Command Line Interface to collecting a summary of the model layers and weights.

    --m: Input model that is necessary to collect to the tensors, this is a
         required *option*
    --v: Verbose: Defines the logging level that the user will be exposed to
    """
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model "
        "and prints names of weights and layers activations "
        "to use with other collect commands",
        prog="ludwig collect_summary",
        usage="%(prog)s [options]",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("collect_summary", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.collect")

    print_ludwig("Collect Summary", LUDWIG_VERSION)

    print_model_summary(**vars(args))


if __name__ == "__main__":
    if len(sys.argv) > 1:
        if sys.argv[1] == "activations":
            cli_collect_activations(sys.argv[2:])
        elif sys.argv[1] == "weights":
            cli_collect_weights(sys.argv[2:])
        elif sys.argv[1] == "names":
            cli_collect_summary(sys.argv[2:])
        else:
            print("Unrecognized command")
    else:
        print("Unrecognized command")



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
from typing import List, Optional, Set

import pandas as pd

from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class SSTLoader(DatasetLoader):
    """The SST dataset.

    This dataset is constructed using the Stanford Sentiment Treebank Dataset.
    This dataset contains binary labels (positive or negative) for each sample.

    The original dataset specified 5 labels:
    very negative, negative, neutral, positive, very positive with
    the following cutoffs:
    [0, 0.2], (0.2, 0.4], (0.4, 0.6], (0.6, 0.8], (0.8, 1.0]
    """

    def __init__(
        self,
        config: DatasetConfig,
        cache_dir: Optional[str] = None,
        include_subtrees=False,
        discard_neutral=False,
        convert_parentheses=True,
        remove_duplicates=False,
    ):
        super().__init__(config, cache_dir=cache_dir)
        self.include_subtrees = include_subtrees
        self.discard_neutral = discard_neutral
        self.convert_parentheses = convert_parentheses
        self.remove_duplicates = remove_duplicates

    @staticmethod
    def get_sentiment_label(id2sent, phrase_id):
        raise NotImplementedError

    def transform_files(self, file_paths: List[str]) -> List[str]:
        # maybe this should be

        """Load dataset files into a dataframe."""
        sentences_df = pd.read_csv(
            os.path.join(self.raw_dataset_dir, "stanfordSentimentTreebank/datasetSentences.txt"),
            sep="\t",
        )

        sentences_df["sentence"] = sentences_df["sentence"].apply(format_text)

        datasplit_df = pd.read_csv(
            os.path.join(self.raw_dataset_dir, "stanfordSentimentTreebank/datasetSplit.txt"), sep=","
        )

        phrase2id = {}
        with open(os.path.join(self.raw_dataset_dir, "stanfordSentimentTreebank/dictionary.txt")) as f:
            Lines = f.readlines()
            for line in Lines:
                if line:
                    split_line = line.split("|")
                    phrase = split_line[0]
                    phrase2id[phrase] = int(split_line[1])

        id2sent = {}
        with open(os.path.join(self.raw_dataset_dir, "stanfordSentimentTreebank/sentiment_labels.txt")) as f:
            Lines = f.readlines()
            for line in Lines:
                if line:
                    split_line = line.split("|")
                    try:
                        id2sent[int(split_line[0])] = float(split_line[1])
                    except ValueError:
                        pass

        trees_pointers = None
        trees_phrases = None

        if self.include_subtrees:
            trees_pointers = []
            with open(os.path.join(self.raw_dataset_dir, "stanfordSentimentTreebank/STree.txt")) as f:
                Lines = f.readlines()
                for line in Lines:
                    if line:
                        trees_pointers.append([int(s.strip()) for s in line.split("|")])

            trees_phrases = []
            with open(os.path.join(self.raw_dataset_dir, "stanfordSentimentTreebank/SOStr.txt")) as f:
                Lines = f.readlines()
                for line in Lines:
                    if line:
                        trees_phrases.append([s.strip() for s in line.split("|")])

        splits = {"train": 1, "test": 2, "dev": 3}

        generated_csv_filenames = []
        for split_name, split_id in splits.items():
            sentence_idcs = get_sentence_idcs_in_split(datasplit_df, split_id)

            pairs = []
            if split_name == "train" and self.include_subtrees:
                phrases = []
                for sentence_idx in sentence_idcs:
                    # trees_pointers and trees_phrases are 0 indexed
                    # while sentence_idx starts from 1
                    # so we need to decrease sentence_idx value
                    sentence_idx -= 1
                    subtrees = sentence_subtrees(sentence_idx, trees_pointers, trees_phrases)

                    sentence_idx += 1
                    sentence_phrase = list(sentences_df[sentences_df["sentence_index"] == sentence_idx]["sentence"])[0]

                    sentence_phrase = convert_parentheses(sentence_phrase)
                    label = self.get_sentiment_label(id2sent, phrase2id[sentence_phrase])
                    # filter @ sentence level
                    # For SST-2, check subtrees only if sentence is not neutral
                    if not self.discard_neutral or label != -1:
                        for phrase in subtrees:
                            label = self.get_sentiment_label(id2sent, phrase2id[phrase])
                            if not self.discard_neutral or label != -1:
                                if not self.convert_parentheses:
                                    phrase = convert_parentheses_back(phrase)
                                    phrase = phrase.replace("\xa0", " ")
                                pairs.append([phrase, label])
            else:
                phrases = get_sentences_with_idcs(sentences_df, sentence_idcs)
                for phrase in phrases:
                    phrase = convert_parentheses(phrase)
                    label = self.get_sentiment_label(id2sent, phrase2id[phrase])
                    if not self.discard_neutral or label != -1:
                        if not self.convert_parentheses:
                            phrase = convert_parentheses_back(phrase)
                            phrase = phrase.replace("\xa0", " ")
                        pairs.append([phrase, label])

            final_csv = pd.DataFrame(pairs)
            final_csv.columns = ["sentence", "label"]
            if self.remove_duplicates:
                final_csv = final_csv.drop_duplicates(subset=["sentence"])
            csv_filename = os.path.join(self.raw_dataset_dir, f"{split_name}.csv")
            generated_csv_filenames.append(csv_filename)
            final_csv.to_csv(csv_filename, index=False)

        return super().transform_files(generated_csv_filenames)


class SST2Loader(SSTLoader):
    """The SST2 dataset.

    This dataset is constructed using the Stanford Sentiment Treebank Dataset.
    This dataset contains binary labels (positive or negative) for each sample.

    The original dataset specified 5 labels:
    very negative, negative, neutral, positive, very positive with
    the following cutoffs:
    [0, 0.2], (0.2, 0.4], (0.4, 0.6], (0.6, 0.8], (0.8, 1.0]

    In the construction of this dataset, we remove all neutral phrases
    and assign a negative label if the original rating falls
    into the following range: [0, 0.4] and a positive label
    if the original rating is between (0.6, 1.0].
    """

    def __init__(
        self,
        config: DatasetConfig,
        cache_dir: Optional[str] = None,
        include_subtrees=False,
        convert_parentheses=True,
        remove_duplicates=False,
    ):
        super().__init__(
            config,
            cache_dir=cache_dir,
            include_subtrees=include_subtrees,
            discard_neutral=True,
            convert_parentheses=convert_parentheses,
            remove_duplicates=remove_duplicates,
        )

    def get_sentiment_label(self, id2sent, phrase_id):
        sentiment = id2sent[phrase_id]
        if sentiment <= 0.4:  # negative
            return 0
        elif sentiment > 0.6:  # positive
            return 1
        return -1  # neutral


class SST3Loader(SSTLoader):
    """The SST3 dataset.

    This dataset is constructed using the Stanford Sentiment Treebank Dataset.
    This dataset contains five labels (very negative, negative, neutral,
    positive, very positive) for each sample.

    In the original dataset, the  5 labels: very negative, negative, neutral, positive,
    and very positive have the following cutoffs:
    [0, 0.4], (0.4, 0.6], (0.6, 1.0]

    This class pulls in an array of mixins for different types of functionality
    which belongs in the workflow for ingesting and transforming
    training data into a destination dataframe that can be use by Ludwig.
    """

    def __init__(
        self,
        config: DatasetConfig,
        cache_dir: Optional[str] = None,
        include_subtrees=False,
        convert_parentheses=True,
        remove_duplicates=False,
    ):
        super().__init__(
            config,
            cache_dir=cache_dir,
            include_subtrees=include_subtrees,
            convert_parentheses=convert_parentheses,
            remove_duplicates=remove_duplicates,
        )

    def get_sentiment_label(self, id2sent, phrase_id):
        sentiment = id2sent[phrase_id]
        if sentiment <= 0.4:
            return "negative"
        elif sentiment <= 0.6:
            return "neutral"
        elif sentiment <= 1.0:
            return "positive"
        return "neutral"


class SST5Loader(SSTLoader):
    """The SST5 dataset.

    This dataset is constructed using the Stanford Sentiment Treebank Dataset.
    This dataset contains five labels (very negative, negative, neutral,
    positive, very positive) for each sample.

    In the original dataset, the  5 labels: very negative, negative, neutral, positive,
    and very positive have the following cutoffs:
    [0, 0.2], (0.2, 0.4], (0.4, 0.6], (0.6, 0.8], (0.8, 1.0]

    This class pulls in an array of mixins for different types of functionality
    which belongs in the workflow for ingesting and transforming
    training data into a destination dataframe that can be use by Ludwig.
    """

    def __init__(
        self,
        config: DatasetConfig,
        cache_dir: Optional[str] = None,
        include_subtrees=False,
        convert_parentheses=True,
        remove_duplicates=False,
    ):
        super().__init__(
            config,
            cache_dir=cache_dir,
            include_subtrees=include_subtrees,
            convert_parentheses=convert_parentheses,
            remove_duplicates=remove_duplicates,
        )

    def get_sentiment_label(self, id2sent, phrase_id):
        sentiment = id2sent[phrase_id]
        if sentiment <= 0.2:
            return "very_negative"
        elif sentiment <= 0.4:
            return "negative"
        elif sentiment <= 0.6:
            return "neutral"
        elif sentiment <= 0.8:
            return "positive"
        elif sentiment <= 1.0:
            return "very_positive"
        return "neutral"


def format_text(text: str):
    """Formats text by decoding into utf-8."""
    return " ".join([w.encode("latin1").decode("utf-8") for w in text.strip().split(" ")])


def convert_parentheses(text: str):
    """Replaces -LRB- and -RRB- tokens present in SST with ( and )"""
    return text.replace("-LRB-", "(").replace("-RRB-", ")")


def convert_parentheses_back(text: str):
    """Replaces ( and ) tokens with -LRB- and -RRB-"""
    return text.replace("(", "-LRB-").replace(")", "-RRB-")


def get_sentence_idcs_in_split(datasplit: pd.DataFrame, split_id: int):
    """Given a dataset split is (1 for train, 2 for test, 3 for dev), returns the set of corresponding sentence
    indices in sentences_df."""
    return set(datasplit[datasplit["splitset_label"] == split_id]["sentence_index"])


def get_sentences_with_idcs(sentences: pd.DataFrame, sentences_idcs: Set[int]):
    """Given a set of sentence indices, returns the corresponding sentences texts in sentences."""
    criterion = sentences["sentence_index"].map(lambda x: x in sentences_idcs)
    return sentences[criterion]["sentence"].tolist()


def sentence_subtrees(sentence_idx, trees_pointers, trees_phrases):
    tree_pointers = trees_pointers[sentence_idx]
    tree_phrases = trees_phrases[sentence_idx]
    tree = SSTTree(tree_pointers, tree_phrases)
    return tree.subtrees()


def visit_postorder(node, visit_list):
    if node:
        visit_postorder(node.left, visit_list)
        visit_postorder(node.right, visit_list)
        visit_list.append(node.val)


class SSTTree:
    class Node:
        def __init__(self, key, val=None):
            self.left = None
            self.right = None
            self.key = key
            self.val = val

    def create_node(self, parent, i):
        if self.nodes[i] is not None:
            # already created
            return
        self.nodes[i] = self.Node(i)

        if parent[i] == -1:
            # is root
            self.root = self.nodes[i]
            return

        if self.nodes[parent[i]] is None:
            # parent not yet created
            self.create_node(parent, parent[i])

        # assign current node to parent
        parent = self.nodes[parent[i]]
        if parent.left is None:
            parent.left = self.nodes[i]
        else:
            parent.right = self.nodes[i]

    def create_tree(self, parents, tree_phrases):
        n = len(parents)
        self.nodes = [None for i in range(n)]
        self.root = [None]
        for i in range(n):
            self.create_node(parents, i)
        for i, phrase in enumerate(tree_phrases):
            self.nodes[i].val = phrase
        for node in self.nodes:
            if node.val is None:
                node.val = " ".join((node.left.val, node.right.val))

    def __init__(self, tree_pointers, tree_phrases):
        self.create_tree([int(elem) - 1 for elem in tree_pointers], tree_phrases)

    def subtrees(self):
        visit_list = []
        visit_postorder(self.root, visit_list)
        return visit_list
